#!/usr/bin/python3

# Routine to plot BMI nomogram
# See derivations of equations at end of code

from pyx import *
import math

def main():


    # Opening boilerplate
    # ====================
    c = canvas.canvas()

    # Transformation function to be used later
    # =========================================
    A = -1906.295/1530 
    B = 0 
    C = 2
    D = 9857.79/1530
    E = 17/90
    F = -50/9
    G = -1488.35/1530
    H = 0    
    tfm = lambda x,y: [(A*x+B*y+C)/(G*x+H*y+1),(D*x+E*y+F)/(G*x+H*y+1)]

    # Plot loci
    # ==========

    # Mass locus(M), M = 40kg to 130kg

    # - locus itself
    p = path.line(*tfm(0,40),*tfm(0,130))
    c.stroke(p)

    # - locus title
    labpos =  [tfm(0,130)[0], tfm(0,130)[1] + .75]
    c.text(*labpos, "Mass (kg)", [text.halign.left, text.valign.top])

    # - major tics and labels
    for ticval in range(40,131,5):
        ticx,ticy = tfm(0,ticval)[0], tfm(0,ticval)[1]
        p = path.line(ticx,ticy,ticx+.3,ticy)
        c.stroke(p)
        c.text(ticx + .5, ticy, str(ticval), 
               [text.halign.boxleft,text.halign.flushright,text.valign.middle])

    # - minor tics
    for ticval in range(40,131):
        ticx,ticy = tfm(0,ticval)[0], tfm(0,ticval)[1]
        p = path.line(ticx,ticy,ticx+.15,ticy)
        c.stroke(p)    

    # BMI locus (B/(B+1)), B from 40kg/2.1m^2 to 130kg/1.4m^2 
    # All tics and labels rotated to be at right angles
    # to the locus.

    # - locus itself
    b_start = tfm(40/(40+2.1**2),0)
    b_end = tfm(130/(130+1.4**2),0)
    ticang = math.atan2(b_end[1]-b_start[1], b_end[0]-b_start[0]) - math.pi/2
    p = path.line(*b_start,*b_end)
    c.stroke(p)

    # - locus title
    labpos =  [b_end[0]-math.sin(ticang)*.6, b_end[1]+math.cos(ticang)*.6]
    c.text(*labpos, "BMI", [text.halign.right, text.valign.top,
                            trafo.rotate(math.degrees(ticang))])

    # - major tics and labels
    for ticval in range(10,66,5):
        ticx,ticy = tfm(ticval/(1+ticval),0)[0], tfm(ticval/(1+ticval),0)[1]
        ticx2,ticy2 = ticx-.3*math.cos(ticang),ticy-.3*math.sin(ticang)
        ticx3,ticy3 = ticx-.5*math.cos(ticang),ticy-.5*math.sin(ticang)
        p = path.line(ticx,ticy,ticx2,ticy2)
        c.stroke(p)
        c.text(ticx3, ticy3, ticval,
               [text.halign.boxright,text.halign.flushleft,
               text.valign.middle,trafo.rotate(math.degrees(ticang))])

    # - minor tics and labels
    for ticval in range(10,66):
        ticx,ticy = tfm(ticval/(1+ticval),0)[0], tfm(ticval/(1+ticval),0)[1]
        ticx2,ticy2 = ticx-.15*math.cos(ticang),ticy-.15*math.sin(ticang)
        p = path.line(ticx,ticy,ticx2,ticy2)
        c.stroke(p)

    # Height locus (-H^2), H = 1.4m to 2.1m

    # - locus itself
    p = path.line(*tfm(1,-(1.4**2)),*tfm(1,-(2.1**2)))
    c.stroke(p)

    # - locus title
    labpos =  [tfm(1,-(1.4**2))[0], tfm(1,-(1.4**2))[1] + .75]
    c.text(*labpos, "Height (m)", [text.halign.right, text.valign.top])

    # - major tics and labels
    for ticval in [i/100 for i in range(140,211,5)]:
        ticx,ticy = tfm(1,-(ticval**2))[0], tfm(1,-(ticval**2))[1]
        p = path.line(ticx,ticy,ticx-.3,ticy)
        c.stroke(p)
        c.text(ticx - .5, ticy, "%3.2f" % (ticval,),
               [text.halign.boxright,text.halign.flushleft,text.valign.middle])

    # - minor tics
    for ticval in [i/100 for i in range(140,211)]:
        ticx,ticy = tfm(1,-(ticval**2))[0], tfm(1,-(ticval**2))[1]
        p = path.line(ticx,ticy,ticx-.15,ticy)
        c.stroke(p)
    
    # Add example of 70kg, 1.68m person
    # ==================================
    p = path.line(*tfm(0,70),*tfm(1,-(1.68**2)))
    c.stroke(p, [color.rgb.red])
    b_example = 70/(1.68**2)
    p = path.circle(*tfm(b_example/(b_example+1),0),.3)
    c.stroke(p, [color.rgb.red])
    
    # Closing boilerplate
    # ===================
    pg = document.page(c,rotated=True,
                       paperformat=document.paperformat.A4)
    doc = document.document([pg]);
    doc.writePDFfile("bmi")

if __name__ == "__main__":
    main()

# Derivations
# ============

# Begin with BMI formula
    # B = M / H^2 => B x H^2 - M = 0
    # Re-express in determinant form with one variable
    # per row and ones in the right column...
    # B x H^2 - M = 0  
    # <=> H^2 x B + B x M - M(B+1) = 0
    # <=> H^2 x B/(B+1) + M x B/(B+1) - M = 0   (B != -1)
    # <=> |    0         M      1  |
    #     |    1       -H^2     1  | = 0
    #     |  B/(B+1)     0      1  |
    # (This is actually a different equation, but the algebra
    # shows that it is true if and only if the original equation is 
    # true, providing B is not -1, which it shouldn't be.)
    #

# With all points plotted it would look a bit like this:
    #
    # M
    # |
    # |
    # |__________________________x=1
    # 0             B             |
    #                             |
    #                             |
    #                             H (y coordinate is -H^2)
    #
    # But this form is quite impractical.

# We can make it more useful by doing a projective transformation, 
    # since collinear points remain collinear under such a 
    # transformation.  
    # Given any four points ABCD (no three collinear), 
    # there exists a unique projective transformation that moves these to 
    # four new points A'B'C'D'.
    # We will plot heights from 1.40 to 2.10m and masses from 
    # 40 to 130kg.  It would be 
    # nice if we could fit this into a nice rectangle about equal
    # to the size of an A4 page minus margins. So we take the initial 
    # coordinate (0,40) to the new coordinate (2,2), etc.
    #
    #  INITIAL  |       NEW
    # x      y  | x'(cm)    y'(cm)
    # - -------- ---------- ------
    # 0     40  |   2.0       2.0 
    # 0    130  |   2.0      19.0 
    # 1 -140^2  |  27.7      19.0
    # 1 -210^2  |  27.7       2.0
    #
    # Homogeneous coordinates are used, and then the projective 
    # transformation is:
    #   /x'\   / A B C \ /x\
    # r |y'| = | D E F | |y|
    #   \1 /   \ G H 1 / \1/
    # 
    # => rx' = Ax + By + C    (i)
    #    ry' = Dx + Ey + F   (ii)
    #    r   = Gx + Gy + 1  (iii)
    # 
    # Substituting (iii) into (i) and (ii) gives
    #
    # Ax + By + C = Gxx' + Hyx' + x'    (iv)
    # Dx + Ey + F = Gxy' + Hyy' + y'     (v)
    #
    # Plugging values for each of our four old 
    # and new points gives eight equations:
    #
    # 0A +     40B + C =    2*0G +       2*40H +  2.0 (1)
    # 0A +    130B + C =    2*0G +      2*130H +  2.0 (2) 
    # 1A - 1.4^2*B + C = 27.7*1G - 1.4^2*27.7H + 27.7 (3)
    # 1A - 2.1^2*B + C = 27.7*1G - 2.1^2*27.7H + 27.7 (4)
    # 0D +     40E + F =    2*0G +       2*40H +  2.0 (5)
    # 0D +    130E + F =   19*0G +     19*130H + 19.0 (6)
    # 1D - 1.4^2*E + F =   19*1G - 1.4^2*19.0H + 19.0 (7)
    # 1D - 2.1^2*E + F =    2*1G -  2.1^2*2.0H +  2.0 (8)
    #
    # From (1) and (2), B = 2H, while from (3) and (4),
    # B =  27.7H, so it must be that B = H = 0, which 
    # implies that C = 2 and that 
    #
    # A = 27.7G + 27.7 - 2 = 27.7G + 25.7             (9)
    # 
    # From (5) and (6), E = 17/90, which implies that 
    # F = 2 - 40E = -50/9.
    #
    # From (7) and (8), G = (2.45E - 17)/17 = -1488.35/1530.
    # Plugging this into (9) gives A = -1906.295/1530.
    #
    # Finally from (8), D = 2G + 2 - 2.1^2*E - F = 9857.79/1530.
    # 


